
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/time.h>		/* needed for gettimeofday */
#include <sys/ioctl.h>		/* needed for ioctl - get tty columns */
#include <dirent.h>			/* needed for scandir */
#include <libgen.h>			/* needed for basename */
#include <stdint.h>			/* needed for uint32_t */

#ifdef HAVE_GETTEXT
#include <libintl.h>
#define _(s) (gettext(s))
#else
#define _(s) (s)
#endif	/* HAVE_GETTEXT */

#include "sequence.h"

extern char *optarg;
extern int optind, opterr, optopt;

extern int errno;

static const char *program_name;

struct options {
	bool interactive : 1;
	bool recursive : 1;
	bool verbose : 1;
	bool failfast : 1;
	bool quiet : 1;
	unsigned int width;
	size_t buf_size;
	char pref_unit;
} static opts = { false, false, false, true, false, 80, 8 * 1024, '\0' };

#define PRINT_BUF_SIZE 256
static char print_buf[PRINT_BUF_SIZE];

char *format_size(off_t s) {
	static const char unit[] = { 'B', 'K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y' };
	int i;
	long double d = s;
	for (i = 0; i < sizeof(unit) / sizeof(unit[0]) && unit[i] != opts.pref_unit && d >= 10240.0; i++)
		d /= 1024.0;
	if (i) {
		snprintf(print_buf, PRINT_BUF_SIZE, "%.1Lf %c", d, unit[i]);
	} else {
		snprintf(print_buf, PRINT_BUF_SIZE, "%llu %c", s, unit[i]);
	}
	return print_buf;
}

void usage(int code) {
	printf("%s %s\n", program_name, _("[-B SIZE] [-i] [-r] [-v] { -t DESTDIR SOURCE... | SOURCE... DEST }"));
	puts("");
	puts(_("\
Copies SOURCE to DEST or SOURCES into DEST(DIR). This program is only capable of\n\
copying regular files and directory structures. However even for regular files\n\
there is the restriction that sparse-files are not supported (yet)."));
	puts("");
	puts(_("Supported standard cp options:"));
//	puts(_("  -i       interactive mode, asks before overwriting files"));
	puts(_("  -r, -R   copy directories recursively"));
	puts(_("  -t DIR   specifies the directory to copy the sources into"));
	puts(_("           (differs from standard cp which moves the SOURCES instead)"));
	puts(_("  -T       treat DEST as regular file"));
	puts(_("  -v       verbose mode"));
//	puts(_("  -x       stay on the filesystem cp has been invoked on"));
	puts("");
	puts(_("Other options:"));
	puts(_("  -B SIZE  buffer size in KByte, default is 8 KByte"));
//	puts(_("  -k       keep going in case of errors"));
	puts(_("  -q       quiet mode"));
	puts(_("  -h       shows this help message"));
	puts("");
	puts(_("\
This program is Free Software and released under the terms of the General Public\n\
License v2 by Franz Brausse (<dev@karlchenofhell.org>) in the hope to be useful."));
	exit(code);
}

#define delta_t int64_t
#define delta(a, b) ((delta_t)((b).tv_sec - (a).tv_sec) * 1000000LL + (delta_t)((b).tv_usec - (a).tv_usec))

#define ERROR(...)	({ fprintf(stderr, __VA_ARGS__); })
#define WARN(...)	({ if (!opts.quiet) fprintf(stderr, __VA_ARGS__); })
#define INFO(...)	({ if (!opts.quiet) printf(__VA_ARGS__); })
#define MSG(...)	({ if (opts.verbose) printf(__VA_ARGS__); })

struct fentry {
	unsigned long blksize;
	char *src;
	char *dest;
	off_t size;
	time_t atime, mtime, ctime;
	uid_t uid;
	gid_t gid;
	mode_t mode;
};

inline static int _log10(uint32_t v) {
	return (
		(v <       10) ? 0 : (v <       100) ? 1 : (v <       1000) ? 2 :
		(v <    10000) ? 3 : (v <    100000) ? 4 : (v <    1000000) ? 5 :
		(v < 10000000) ? 6 : (v < 100000000) ? 7 : (v < 1000000000) ? 8 : 9
	);
}

static uint32_t total_num = 0, proc_num = 0;
static off_t total_size = 0, proc_size = 0;
static bool interrupted = false;
static struct timeval start_time, tmp;

static void print_progress(
		const struct fentry *f,
		const struct timeval *file_start,
		const off_t file_proc) {
	delta_t d_usec;
	double speed_bs, d_size_prog, eta_sec;
	div_t meta, heta;
	
	gettimeofday(&tmp, NULL);
	printf("\r");
	
	// total stats
	d_usec = delta(start_time, tmp);							// delta time in µs
	speed_bs = (proc_size * 1000.0) / (d_usec / 1000.0);		// speed in bytes / sec
	d_size_prog = proc_size / (double)total_size;				// finished in bytes
	// const double d_tnum_prog = total_num / (double)proc_num;	// num files finished
	eta_sec = (total_size - proc_size) / speed_bs;				// eta in sec
	
	int cols = 0;
	// construct format "%?u/%?u%c" where '?' is being replaced by a decimal value
	// permitting enough space for the result to hold total_num as well
	snprintf(print_buf, PRINT_BUF_SIZE, "%%%du%%c", _log10(total_num) + 1);
	cols += printf(print_buf, proc_num, '/');
	cols += printf(print_buf, total_num, ' ');
	
	// print rest
	meta = div(eta_sec, 60);
	heta = div(meta.quot, 60);
	cols += printf(_("total: %5.1f%% ETA %u:%02u:%02u, %s/sec"),
			d_size_prog * 100.0, heta.quot, heta.rem, meta.rem, format_size(speed_bs));
	
	cols += printf(" -- ");
	
	// file stats
	d_usec = delta(*file_start, tmp);							// delta time in µs
	speed_bs = (file_proc * 1000.0) / (d_usec / 1000.0);		// speed in bytes / sec
	d_size_prog = file_proc / (double)f->size;					// finished in bytes
	eta_sec = (f->size - file_proc) / speed_bs;					// eta in sec
	
	meta = div(eta_sec, 60);
	cols += printf(_("%5.1f%% ETA: %02u:%02u, %s/sec"),
			d_size_prog * 100.0, meta.quot, meta.rem, format_size(speed_bs));
	
	snprintf(print_buf, opts.width - cols + 1, " - %s", f->src);
	printf("%s", print_buf);
	fflush(stdout);
}

void free_entry(void *e) {
	free(((struct fentry *)e)->src);
	free(((struct fentry *)e)->dest);
	free(e);
}

#define min(a, b) ((a < b) ? a : b)

bool do_copy(list srcs) {
	char *buf = malloc(opts.buf_size);
	if (!buf) {
		perror(_("Error allocating memory"));
		return false;
	}
	
	bool err = false;
	while (!lst_empty(srcs)) {
		struct fentry *fe = lst_data(lst_first(srcs));
		err = false;
		proc_num++;
		
		if (S_ISDIR(fe->mode)) {
			INFO(_("creating directory '%s'\n"), fe->dest);
			if (mkdir(fe->dest, fe->mode & (S_IRWXU | S_IRWXG | S_IRWXO))) {
				ERROR(_("Error creating directory '%s': %s\n"), fe->dest, strerror(errno));
				err = true;
				goto cont;
			}
			proc_size += fe->size;
			
		} else {
			FILE *fp_src, *fp_dest;
			struct timeval file_start;
			
			MSG("copying '%s' -> '%s'\n", fe->src, fe->dest);
			gettimeofday(&file_start, NULL);
			
			if (!(fp_dest = fopen(fe->dest, "wb"))) {
				err = true;
				ERROR(_("Error opening destination file '%s' for writing: %s\n"), fe->dest, strerror(errno));
				goto cont;
			}
			
			if (!(fp_src = fopen(fe->src, "rb"))) {
				err = true;
				ERROR(_("Error opening source file '%s' for reading: %s\n"), fe->src, strerror(errno));
				goto close_dest;
			}
			
			off_t left = fe->size;
			while (left) {
				size_t read = fread(buf, 1, min(opts.buf_size, left), fp_src);
				if (ferror(fp_src)) {
					err = true;
					perror(_("Error reading source file"));
					goto close_src;
				}
				
				size_t written = fwrite(buf, 1, read, fp_dest);
				if (ferror(fp_dest)) {
					err = true;
					perror(_("Error writing destination file"));
					goto close_src;
				}
				
				if (read != written) {
					ERROR(_("unequal amount bytes tranferred: %d read, %d written\n"), read, written);
				}
				
				left -= read;
				proc_size += read;
				
				print_progress(fe, &file_start, fe->size - left);
				
				if (interrupted) {
					err = true;
					break;
				}
			}
			
			puts("");
			
			// todo: set ownership / mode
			
			close_src:
			if (fclose(fp_src) == EOF) {
				err = true;
				ERROR(_("Error closing source file '%s': %s\n"), fe->src, strerror(errno));
			}
			close_dest:
			if (fclose(fp_dest) == EOF) {
				err = true;
				ERROR(_("Error closing destination file '%s': %s\n"), fe->dest, strerror(errno));
			}
		}
		
		cont:
		if (err && opts.failfast)
			break;
		free_entry(lst_pop(srcs));
	}
	
	free(buf);
	return !err;
}

static char * path_concat(const char *a, const size_t a_len, const char *b, const size_t b_len) {
	const bool needs_slash = (a[a_len - 1] != '/');
	size_t off = 0;
	char *dest_path = malloc(a_len + needs_slash + b_len + 1);
	if (!dest_path) {
		perror(_("Error allocating memory"));
		return NULL;
	}
	strncpy(dest_path, a, a_len);
	off += a_len;
	if (needs_slash)
		dest_path[off++] = '/';
	strncpy(dest_path + off, b, b_len);
	off += b_len;
	dest_path[off] = '\0';
	return dest_path;
}

bool check_dir(list srcs, const struct fentry *fe) {
	struct dirent **dirs;
	
	int n = scandir(fe->src, &dirs, 0, alphasort);
	if (n < 0) {
		ERROR(_("Error scanning directory '%s': %s\n"), fe->src, strerror(errno));
		return false;
	}
	
	const char *src = fe->src, *dest = fe->dest;
	const size_t src_len = strlen(src), dest_len = strlen(dest);
	
	while (n--) {
		const char *d_name = dirs[n]->d_name;
		const size_t len = strlen(d_name);
		switch (len) {
			case 1:
			case 2:	// break when dirs[n] is '.' or '..'
				if (!strncmp("..", d_name, len))
					break;
				// fall-through otherwise
			default: {
				struct fentry *e = malloc(sizeof(struct fentry));
				e->src = path_concat(src, src_len, d_name, len);
				e->dest = path_concat(dest, dest_len, d_name, len);
				MSG("in dir: d_name '%s', src '%s', dest '%s'\n", d_name, e->src, e->dest);
				lst_push(srcs, e);
			} break;
		}
		free(dirs[n]);
	}
	free(dirs);
	return true;
}

bool check_sources(list srcs, list l) {
	struct stat s, d;
	
	while (!interrupted && !lst_empty(srcs)) {
		struct fentry *fe = lst_pop(srcs);
		
		if (lstat(fe->src, &s)) {
			ERROR(_("Error accessing file '%s': %s\n"), fe->src, strerror(errno));
			goto err_cleanup;
		}
		
		const char *type = NULL;
		const char *expl = NULL;
		switch (s.st_mode & S_IFMT) {
			case S_IFSOCK:
				type = "socket";
				break;
			case S_IFLNK:
				type = "symbolic link";
				break;
			case S_IFREG: // regular file
				if (s.st_blocks < (s.st_size + 511) >> 9) {
					type = "sparse file";
					if (opts.verbose) {
						char *size = strdup(format_size(s.st_size));
						snprintf(print_buf, PRINT_BUF_SIZE,
								_("file size: %s => 512B-blocks: %llu, actual blocks: %llu"),
								size, (s.st_size + 511) >> 9, s.st_blocks);
						free(size);
					} else {
						const double holes = (s.st_size - (s.st_blocks << 9)) / (double)s.st_size;
						snprintf(print_buf, PRINT_BUF_SIZE, _("up to %.1f%% holes in file"), holes * 100.0);
					}
					expl = print_buf;
				}
				break;
			case S_IFBLK:
				type = "block device";
				break;
			case S_IFDIR: {
				if (!opts.recursive) {
					type = "directory";
					break;
				}
				if (lstat(fe->dest, &d)) {
					if (errno != ENOENT) {
						type = "directory";
						snprintf(print_buf, PRINT_BUF_SIZE,
								_("error accessing destination '%s': %s"),
								fe->dest, strerror(errno));
						expl = print_buf;
						break;
					}
				} else if (!S_ISDIR(d.st_mode)) {
					type = "directory";
					snprintf(print_buf, PRINT_BUF_SIZE,
							_("destination '%s' exists and is not a directory"),
							fe->dest);
					expl = print_buf;
					break;
				}
				
				if (!check_dir(srcs, fe))
					goto err_cleanup;
			} break;
			case S_IFCHR:
				type = "character device";
				break;
			case S_IFIFO:
				type = "named pipe";
				break;
			default:
				type = "unknown file-type";
				snprintf(print_buf, PRINT_BUF_SIZE, "mode: %o", s.st_mode);
				expl = print_buf;
				break;
		}
		
		if (type) {
			if (expl) {
				WARN(_("Warning: won't copy %s '%s': %s\n"), _(type), fe->src, expl);
			} else {
				WARN(_("Warning: won't copy %s '%s'\n"), _(type), fe->src);
			}
			goto err_cleanup;
		}
		
		fe->blksize = s.st_blksize;
		fe->size = s.st_size;
		fe->atime = s.st_atime;
		fe->mtime = s.st_mtime;
		fe->ctime = s.st_ctime;
		fe->mode = s.st_mode;
		fe->gid = s.st_gid;
		fe->uid = s.st_uid;
		
		lst_push(l, fe);
		
		MSG("copying '%s' (%llu bytes)\n", fe->src, fe->size);
		
		total_size += s.st_size;
		total_num++;
		continue;
		
err_cleanup:
		free_entry(fe);
		if (opts.failfast)
			return false;
	}
	
	return true;
}

bool parse_winsize_ioctl() {
	struct winsize ws;
	if (!ioctl(STDOUT_FILENO, TIOCGWINSZ, &ws) && ws.ws_col > 0) {
		opts.width = ws.ws_col;
		return true;
	}
	return false;
}

void parse_environ(char **);

void handler(int signal) {
	switch (signal) {
		case SIGINT:
			MSG(_("Interrupted by SIGINT\n"));
			interrupted = true;
			break;
		case SIGWINCH:
			parse_winsize_ioctl();
			break;
		default:
			psignal(signal, _("unknown signal"));
			break;
	}
}

int main(int argc, char *argv[], char **envp) {
	program_name = argv[0];
	
	parse_environ(envp);
	
	char *dest = NULL;
	bool in_dest = false, in_dest_set = false;
	
	int opt;
	while ((opt = getopt(argc, argv, "B:ikqrRt:Tvh")) != -1) {
		switch (opt) {
			case 'B':
				opts.buf_size = atoi(optarg);
				break;
			case 'i':
				opts.interactive = true;
				break;
			case 'k':
				opts.failfast = false;
				break;
			case 'q':
				opts.quiet = true;
				break;
			case 'r':
			case 'R':
				opts.recursive = true;
				break;
			case 't':
				dest = optarg;
				in_dest = true;
				in_dest_set = true;
				break;
			case 'T':
				in_dest = false;
				in_dest_set = true;
				break;
			case 'v':
				opts.verbose = true;
				break;
			case 'h':
				usage(EXIT_SUCCESS);
				break;
			default: /* '?' */
				exit(EXIT_FAILURE);
				break;
		}
	}
	
	int srcc = argc - optind;
	if (!dest) {
		dest = argv[argc - 1];
		argv[argc - 1] = NULL;
		srcc--;
	}
	
	if (srcc <= 0) {
		ERROR(_("no sources given\n"));
		exit(EXIT_FAILURE);
	}
	
	// remove trailing slash of dest
	size_t dest_len = strlen(dest);
	if (dest[dest_len - 1] == '/')
		dest[--dest_len] = '\0';
	
	struct stat s;
	if (lstat(dest, &s)) {
		if (errno == ENOENT) {
			if (in_dest_set) {
				if (in_dest) {
					// create directory dest
					if (mkdir(dest, S_IRWXU | S_IRWXG | S_IRWXO)) {
						ERROR(_("Error creating destination directory '%s': %s\n"), dest, strerror(errno));
						exit(EXIT_FAILURE);
					} else {
						MSG(_("Created destination directory '%s'\n"), dest);
					}
				}
			} else {
				// todo: make dependant on (S_ISDIR(srcs[0]) || srcc > 1)
				in_dest = false;
			}
		} else {
			ERROR(_("Error stat'ing destination '%s': %s\n"), dest, strerror(errno));
			exit(EXIT_FAILURE);
		}
	} else {
		bool is_dir = S_ISDIR(s.st_mode);
		if (in_dest_set) {
			if (in_dest && !is_dir) {
				ERROR(_("Error: destination '%s' is not a directory\n"), dest);
				exit(EXIT_FAILURE);
			} else if (!in_dest && is_dir) {
				ERROR(_("Error: destination '%s' is not a file\n"), dest);
				exit(EXIT_FAILURE);
			}
		} else {
			in_dest = is_dir;
		}
	}
	
	char **srcs = argv + optind;
	
	if (!in_dest) {
		if (srcc == 1) {
			if (lstat(srcs[0], &s)) {
				ERROR(_("Error stat'ing source '%s': %s\n"), srcs[0], strerror(errno));
				exit(EXIT_FAILURE);
			} else if (S_ISDIR(s.st_mode)) {
				ERROR(_("Error: source '%s' is a directory\n"), srcs[0]);
				exit(EXIT_FAILURE);
			}
		} else /* srcc > 1 */ {
			ERROR(_("Error: destination '%s' is not a directory\n"), dest);
			exit(EXIT_FAILURE);
		}
	}
	
	void *org_sigint_handler = signal(SIGINT, handler);
	void *org_sigwinch_handler = signal(SIGWINCH, handler);
	
	list l, d;
	lst_init(&l);
	lst_init(&d);
	int i;
	
	for (i = 0; i < srcc; i++) {
		char *src = strdup(srcs[i]);
		const char *bname = basename(srcs[i]);
		char *_dest = (in_dest) ? path_concat(dest, dest_len, bname, strlen(bname)) : strdup(dest);
		
		// todo: check whether src and dest are the same file
		
		struct fentry *e = malloc(sizeof(struct fentry));
		e->src = src;
		e->dest = _dest;
		
		MSG("destination: for '%s' (%s) (%d): %s\n", e->src, bname, in_dest, e->dest);
		lst_push(l, e);
	}
	
	if (!check_sources(l, d))
		goto cleanup;
	if (interrupted)
		goto cleanup;
	
	INFO(_("total: %u files, %s\n"), total_num, format_size(total_size));
	
	gettimeofday(&start_time, NULL);
	if (!do_copy(d))
		ERROR(_("Error copying file\n"));
	
	cleanup:
	lst_free_data(&l, free_entry);
	lst_free_data(&d, free_entry);
	
 	signal(SIGINT, org_sigint_handler);
 	signal(SIGWINCH, org_sigwinch_handler);
	
	return EXIT_SUCCESS;
}

void parse_environ(char **envp) {
	if (parse_winsize_ioctl())
		return;
	
	char *cc;
	while ((cc = *envp++)) {
		if (strstr(cc, "COLUMNS=")) {
			errno = 0;
			char *err = NULL;
			unsigned long int v = strtoul(cc + strlen("COLUMNS="), &err, 10);
			if (errno) {
				WARN(_("Error parsing environment variable COLUMNS at %c: %s\n"), *err, strerror(errno));
				break;
			}
			if (!v)
				break;
			opts.width = v;
			return;
		}
	}
	
	const char *col_str = getenv("COLUMNS");
	if (col_str) {
		errno = 0;
		char *err = NULL;
		unsigned long int v = strtoul(col_str, &err, 10);
		if (errno) {
			WARN(_("Error parsing environment variable COLUMNS at %c: %s\n"), *err, strerror(errno));
		} else if (v) {
			opts.width = v;
			return;
		}
	}
}
